#
# Compute the number of rings for a sensors with
# the given angular resolution and radius R
#

tm.logPolarRings <- function(R, angularResolution = 72) {
  #
  pr <- 2 * acos(-1) * R / angularResolution
  a  <- (R-pr)/R
  #
  return(as.integer( 1-log(R) / log(a)))
  #
}

#
# Compute the log polar coordinates of the pixel
# with given cartesian coordinates for a sensor
# of radius R and given angularResolution
#
#

tm.logPolarCoordinates <- function(x, y, R, angularResolution = 72) {
  #
  r      <- sqrt(x*x+y*y)
  theta  <- atan2(y,x) 
  dtheta <- (2*acos(-1)) / angularResolution
  #
  if(theta < 0)
    theta <- theta + 2*acos(-1)
  #
  theta <- dtheta * (theta %/% dtheta)
  #
  # compute the linear size of a pixel 
  # at distance R
  #
  pr <- 2 * acos(-1) * R / angularResolution
  #
  # and use this information to compute
  # the scaling factor from one ring of pixels
  # to next (inner) one
  #
  a  <- (R-pr)/R
  #
  # The ring are numbered from 0 starting from the
  # outer ring: if you need to start from the
  # smallest ring you should complement the (first)
  # coordinate with the number of rings:
  # i -> (#rings - 1) - i 
  #
  i <- (log(r) - log(R)) %/% log(a)
  #
  c(as.integer(i), as.integer((theta %/% dtheta)))
  #
}

#
# Compute the (cartesian) coordinates of the
# point that will be sampled in the process of building the
# log polar mapping for a sensor of radius R
# and a given angularResolution.
#

tm.logPolarSampleCoordinates <- function(x, y, R, angularResolution = 72) {

  r      <- sqrt(x*x+y*y)
  theta  <- atan2(y,x) 
  dtheta <- (2*acos(-1)) / angularResolution
    
  if(theta < 0)
    theta <- theta + 2*acos(-1)

  theta <- dtheta * (theta %/% dtheta)

  pr <- 2 * acos(-1) * R / angularResolution
  a  <- (R-pr)/R

  i <- (log(r) - log(R)) %/% log(a)

  rr <- exp(i*log(a)+log(R))

  c(rr*cos(theta), rr*sin(theta))
  
}

#
# Map the log polar coordinates (ring number and angular position)
# into cartesian coordinates.
#

tm.logPolarToCartesian <- function(i, itheta, R, angularResolution = 72) {
  #
  dtheta <- (2*acos(-1)) / angularResolution
  theta  <- dtheta * itheta
  #
  pr     <- 2 * acos(-1) * R / angularResolution
  a      <- (R-pr)/R
  rr     <- exp(i*log(a)+log(R))
  #
  c(rr*cos(theta), rr*sin(theta))
  #
}

#
# Generate a log polar map (array) of the given image within the
# bounds of the given R and the specified angular resolution
#

tm.logPolarMap <- function(simg, R, angularResolution = 72) {
  #
  rings   <- tm.logPolarRings(R, angularResolution)
  sectors <- angularResolution
  #
  res     <- array(0, dim = c(sectors, rings))
  data    <- simg@data
  w       <- dim(data)[2]
  h       <- dim(data)[1]
  #
  ox   <- (w+1L)%/%2L
  oy   <- (h+1L)%/%2L
  #
  for (r in 1:rings) {
    for (s in 1:sectors) {
      xy <- tm.logPolarToCartesian(r, s, R, angularResolution);
      res[s, (rings-r) + 1] <- data[xy[2]+oy, xy[1]+ox] 
    }
  }
  #
  res
  #
}

#
# The spatial non uniformity of log polar mapping
# mandates for appropriate space variant smoothing. 
# Our approach is based on the use of progreesive smoothing by
# the heat equation, building the final image by gathering
# the values when they are close to the space variant sigma
# related to the size of each log-polar pixel in the original
# cartesian domain.
# 

tm.logPolarSmoothing <- function(img, angularResolution = 72) {
  #
  # We first need to create a distance map from the origin
  # that we fix in the center of the image
  #
  data <- img@data
  s    <- max(data)
  dist <- array(0, dim(data))
  #
  w    <- dim(data)[2]
  h    <- dim(data)[1]
  #
  ox   <- (w+1L)/2L
  oy   <- (h+1L)/2L
  #
  for (y in 1:h) {
    for (x in 1:w) {
      dx <- x - ox
      dy <- y - oy
      #
      r  <- sqrt(dx*dx + dy*dy)
      #
      dist[y,x] <- r
    }
  }
  #
  # Then we map the distance value into the correct sigma
  #
  dist <- dist * (((2 * acos(-1)) / angularResolution) / sqrt(2))
  #
  message(sprintf("Max required smoothing sigma is %f", (max(dist))))
  #
  # and we employ anisotropic diffusion to smooth the image
  # in a space variant way (in this case more smoothing at the boundary
  #
  res <- ia.mult(tm.spaceVariantDiffusion(ia.div(img, s),
                                          dist,
                                          maxSigma = max(dist), fh = 0.01),
                 s)
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  #
  return(res)
  #
}

#
# Re-render a standard 'cartesian image'  from the information
# available in its log polar version: an appropriately smoothed
# version of the original image must be provided 
#

tm.logPolarMapEquivalent <- function(simg, # the smoothed image
                                     R,
                                     angularResolution =  72,
                                     outside           = 255) {
  #
  data <- simg@data
  res  <- data*0 + outside
  w    <- dim(data)[2]
  h    <- dim(data)[1]
  #
  ox   <- (w+1L)%/%2L
  oy   <- (h+1L)%/%2L
  #
  R <- min(R, ox-1, oy-1)
  #
  for (y in 1:h) {
    for (x in 1:w) {
      dx <- x - ox
      dy <- y - oy
      #
      if((dx != 0) || (dy != 0)) {
        if(sqrt(dx*dx+dy*dy) < R) {
          xy <- tm.logPolarSampleCoordinates(dx, dy, R, angularResolution)
          #
          res[y,x] <- data[round(oy+xy[2]),round(ox+xy[1])]
        }
      } else {
        res[y,x] <- data[y,x]
      }
    }
  }
  #
  res <- new("animage",
             data     = res,
             focus    = simg@focus,
             outside  = simg@outside,
             storage  = "real",
             document = simg@document)
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  #
  return(res)
  #
}
