# The definition of some utility functions
#
# The error function
#
tm.erf  <- function(x) 2 * pnorm(x * sqrt(2)) - 1

#
# The complementary error function
#
tm.erfc <- function(x) 2 * pnorm(x * sqrt(2), lower = FALSE)

#
# and a simple spline approximation to the inverse error function:
# it works well for x <= 0.85, which is fine as we are usually interested in 
# small values (x << 1)
#
tm.ierf <- splinefun(tm.erf(seq(-1, 1, by=0.01)), seq(-1, 1, by=0.01))

#
# Estimate the standard deviation of image noise.
# In the case of zero average Gaussian noise, it completely
# characterizes it.
#
tm.estimateNoiseLevel <- function(img) {
    mad(ia.frame(ia.convolution(img,
                                array(c(0,-1/sqrt(2),1/sqrt(2)),c(1,3))),
                 3)@data)
}

#
# The following function implements applies to the
# argument image the separable basis sets for
# the derivatives of a Gaussian based on
#
# 'The Design and Use of Steerable Filters' Freeman and  Adelson 1991
#
#          order 0: plain smoothing
#          order 1: return the x and y derivative
#          order 2: return G2a, G2b, and G2c
#
# The value of the derivative along a given direction can then be obtained
# as
#
#          order 0: -
#          order 1: cos(t) * G1x + sin(t) * G1y
#          order 2: cos^2(t) G2a - 2 cos(t) sin(t) G2b +sin^2(t) G2c
#
# NOTE 1: As the reference coordinate system we are using in the images is not
# right handed the sign of the derivatives in tm.regularizedDerivatives
# have been modified and their usage in tm.edges adapted to the left handed
# reference system used.
#
# NOTE 2: In order to minimize the effects of boundary conditions the 'outside'
# value of processed images is modified during th eprocess to ensure that no
# spurious derivatives are generated
#
#
tm.regularizedDerivatives <- function(img, order, sigma)
  {
    actualSupport <- as.integer(sigma * 3.5)

    s <- (1/(sqrt(2*pi)*sigma)) 

    if(order == 0) {
      kernel        <- array(0, c(1, 2*actualSupport+1))

      i <- 1
      for(x in -actualSupport:actualSupport) {
        kernel[1, i] <- exp(-(x*x)/(2*sigma*sigma))
        i            <- i+1
      }
      kernel <- kernel / sum(kernel)
      
      rimg <- ia.convolution(ia.convolution(img, kernel),
                             t(kernel))

      return(rimg)
    } else if (order == 1) {
      kernelx        <- array(0, c(1, 2*actualSupport+1))
      kernel         <- array(0, c(1, 2*actualSupport+1))

      i <- 1
      
      for(x in -actualSupport:actualSupport) {
        kernel[1, i]  <-  s * exp(-(x*x)/(2*sigma*sigma))
        kernelx[1, i] <- -s * (-x/(sqrt(2)*sigma)) * exp(-(x*x)/(2*sigma*sigma))
        i             <- i+1
      }

      kernel <- kernel / sum(kernel)

      rimgx <- ia.convolution(ia.convolution(img, kernelx),
                              t(kernel))
      rimgy <- ia.convolution(ia.convolution(img, kernel),
                              t(kernelx))

      
      rimgx          <- ia.convolution(img, kernelx)
      rimgx@outside  <- 0
      rimgx          <- ia.convolution(rimgx, t(kernel))
      
      rimgy          <- ia.convolution(img, -t(kernelx))
      rimgy@outside  <- 0
      rimgy          <- ia.convolution(rimgy, kernel)
      
      return(list(rimgx, rimgy))
    } else if (order == 2) {
      kernel         <- array(0, c(1, 2*actualSupport+1))
      kernelxx       <- array(0, c(1, 2*actualSupport+1))
      kernelxy       <- array(0, c(1, 2*actualSupport+1))

      sxx <- 1/(2*pi*(sigma**4))
      sxy <- sqrt(1/(2*pi*(sigma**6)))
      
      i <- 1
      for(x in -actualSupport:actualSupport) {
        kernel[1, i]   <- exp(-(x*x)/(2*sigma*sigma))
        kernelxx[1, i] <- exp(-(x*x)/(2*sigma*sigma)) * (x*x/(sigma*sigma) - 1) * sxx
        kernelxy[1, i] <- exp(-(x*x)/(2*sigma*sigma)) * x * sxy
        i              <- i+1
      }
        
      rimgxx         <- ia.convolution(img, kernelxx)
      rimgxx@outside <- 0
      rimgxx         <- ia.convolution(rimgxx, t(kernel))
        
      rimgxy         <- ia.convolution(img, kernelxy)
      rimgxy@outside <- 0
      rimgxy         <- ia.convolution(rimgxx, t(kernelxy))
        
      rimgyy         <- ia.convolution(img, t(kernelxx))
      rimgyy@outside <- 0
      rimgyy         <- ia.convolution(rimgyy, kernel)

      list(rimgxx,
           rimgxy,
           rimgyy)
    } else {
      stop("illegal order for derivative")
    }
  }

#
# Compute significant edges based on the 'low resolution' strategy
# outlined in the Elder-Zucker 1998 paper.
#
tm.edgeDetection <- function(img,
                             scale = 1.0,    # smoothing used for regularized derivatives
                             nstd = NULL,    # noise standard deviation
                             alpha = 0.0001, # the lower, the stricter the threshold
                             maxl = 6,       # maximal edge blur for detection
                             onlyFirstThr = FALSE) { # if TRUE, just threshold first deriv
  #
  document     <- img@document
  #
  img@document <- FALSE
  #
  if(img@storage == "integer")
    img <- ia.realAnimage(img)
  #
  if(is.null(nstd)) {
    nstd <- tm.estimateNoiseLevel(img)
  }
  #
  d1   <- tm.regularizedDerivatives(img,1,scale)
  d2   <- tm.regularizedDerivatives(img,2,scale)
  #
  d1x  <- d1[[1]]@data
  d1y  <- d1[[2]]@data
  d2xx <- d2[[1]]@data
  d2xy <- d2[[2]]@data
  d2yy <- d2[[3]]@data
  #
  dirs <- ia.atan2(d1[[2]], d1[[1]])@data
  g    <- ia.sqrt(ia.add(ia.mult(d1[[1]], d1[[1]]),
                         ia.mult(d1[[2]], d1[[2]])))@data
  e    <- array(0, dim=dim(g))
  #
  thr1 <- 0.5 * sqrt(-log(alpha)/pi) * nstd / (scale*scale)
  thr2 <- 0.5 * sqrt(3/(2*pi)) * (nstd/(scale**3)) * tm.ierf(1.0 - alpha)
  #
  r    <- dim(e)[1]
  c    <- dim(e)[2]
  #
  for(y in 1:r) {
    for(x in 1:c) {
      #
      # Take into account distance from boundary
      # necessary when we scan for maxima/minima
      # of the second derivative
      #
      L <- min(min(maxl, min(y-1,r-y)),
               min(maxl, min(x-1,c-x)))
      #
      # if the first derivative is above threshold
      #
      if(g[y,x] > thr1) {
        #
        # if simple thresholding is required
        #
        if(onlyFirstThr == TRUE) {
          #
          # we are done
          #
          e[y,x] <- 1
        } else if (L > 1) {
          #
          # otherwise we explore the gradient line,
          #
          theta <-  dirs[y,x]
          #
          # computing its versor
          #
          ux    <-  cos(theta)
          uy    <- -sin(theta)
          #
          M     <- -1
          lM    <-  0
          left  <- array(0, L)
          right <- array(0, L)
          #
          # we first check for the presence of a significant
          # maximum on one side
          #
          for(l in L:1) {
            yy <- y - uy*l
            xx <- x - ux*l
            gg <- ux*ux*d2xx[yy,xx] + 2*ux*uy*d2xy[yy,xx] + uy*uy*d2yy[yy,xx]
            left[l] <- gg
            if(gg > M) {
              M  <- gg
              lM <- -l
            }
          }
          ll <- gg
          #
          # and, if we found it, we proceed to find a significant
          # minimum on the other side
          #
          if(M > thr2) {
            m  <- 1
            lm <- 0
            for(l in L:1) {
              yy         <- y + uy*l
              xx         <- x + ux*l
              gg         <- ux*ux*d2xx[yy,xx] + 2*ux*uy*d2xy[yy,xx] + uy*uy*d2yy[yy,xx]
              right[l] <- gg
              if(gg < m) {
                m  <- gg
                lm <- l
              }
            }
            lr <- gg
            #
            # we accept it as an edge point if:
            #
            # - the maximum/minimum are equally distant
            # - the sign of the second derivatives at the close left
            #   right of the pixel have different sign
            # - the minimum value is significant (we already checked
            #   the significance of the maximum)
            #
            if((abs(lm+lM) <= 1)      &&
               (ll*lr < 0)            &&
               (m     < -thr2)) {
              e[y,x]         <- lm
            }
          }
        }
      }
    }
  }
  #
  img@document <- document
  #
  res <- animage(e, focus=img@focus, outside=0, storage="real")
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  res
}
