\name{Imaging}

\alias{Imaging}
\alias{tm.computeDistortionMaps}
\alias{tm.distortImage}
\alias{tm.computeVignettingMap}
\alias{tm.vignetteImage}
\alias{tm.Lanczos}

\alias{tm.diffractionKernel}
\alias{tm.photonNoise}
\alias{tm.bayerizeImage}
\alias{tm.debayerizeImage}

\title{Accurate camera simulation support}

\description{ A set of functions designed to operate on synthetic images
  to make them more closely resembling pictures taken with physical devices.
}

\usage{
tm.Lanczos(r, support = 2)

tm.diffractionKernel(focalLength   = 50e-3,
                     fStop         = 11,
                     wavelength    = 550e-9,
                     pixelSize     = 6.4e-6,
                     superSampling = 1)

tm.computeDistortionMaps(width = 301, height = 201,
                         scaling, curve, rmax = 21.6)
tm.distortImage(img, maps)

tm.computeVignettingMap(width = 301, height = 201, 
                        scaling, curve)
tm.vignetteImage(img, map)

tm.photonNoise(x,
               maxValue        =   255,
               gammaCorrection =   (1.0 / 2.2),
               fullWell        = 51400,
               fullWellIso     =   100,
               iso             =   200)

tm.bayerizeImage(rgbImage)
tm.debayerizeImage(img)
}

\arguments{
 \item{curve}{a \code{splinefun} object describing the effect defined over $[0,r_{\max}]$}
 \item{focalLength}{a real representing the focal length of the lens expresse in meters}
 \item{fStop}{a real representing the ratio of lens diameter to the aperture}
 \item{fullWell}{the maximum number of electrons that can be stored in a sensor well (pixel) which correspond to the number of photons captured}
 \item{fullWellIso}{the ISO value at which \code{fullWell} data refers to}
 \item{gammaCorrection}{the non linear mapping applied to the arument image: it must be undone to map to the number of photons}
 \item{height}{an integer representing image (or array) height}
 \item{img}{an \code{animage}}
 \item{iso}{film speed expressed according to linear scale defined by ISO 5800:1987}
 \item{map,maps}{an array, a list of two arrays}
 \item{maxValue}{a real representing the maximum value assumed by pixels in a single image channel (typically 1 or 255)}
 \item{pixelSize}{a real representing the linear size of a pixel (of the
   order of 10e-6)}
 \item{r}{a real representing the radial distance}
 \item{rgbImage}{a list of three \code{animage}s}
 \item{rmax}{reals representing the maximum value for which \code{curve} is defined}
 \item{scaling}{a real scale factor mapping image size into metric sensor size}
 \item{superSampling}{an integer specifying the resolution ratio of the kernel to the image}
 \item{support}{an integer specifying the support radius}
 \item{wavelength}{a real representing the wavelength of light expressed in meters (typically 550e-9)}
 \item{width}{an integer representing image (or array) width}
 \item{x}{a real representing the value of a pixel (for a single image channel)}
}

\value{

  The functions (right) return several kind of values (left):

  \item{real}{\code{tm.photonNoise}, \code{tm.Lanczos}}
  \item{animage}{\code{tm.bayerizeImage},\code{tm.vignetteImage},\code{tm.distortImage}}
  \item{(animage, ...)}{\code{tm.debayerizeImage}}
  \item{array}{\code{tm.computeVignettingMap}, \code{tm.diffractionKernel}}
  \item{(array, ...)}{\code{tm.computeDistortionMaps}}

}

\details{

  The functions  \code{tm.bayerizeImage} \code{tm.debayerizeImage} are
  used to  transform a  full RGB, three  channels image into  a single
  channel image whose pixels represent red, green, or blue information
  depending on their coordinates.
    
  The functions \code{tm.computeVignettingMap} \code{tm.vignetteImage}
  are used respectively to generate a darkening map and to apply it to
  an animage.  The effect is described by a \code{splinefun} reporting
  the fraction  of light ($\leq  1$ that effectively reaches  the frame
  plane (see the  sample curve \code{distagonVignette}).  The vignetting
  is assumed to be radial: the arguments \code{rmin} and \code{rmax}
  identify the extrema  of the vignetting curve (defaulting  to the values
  appropriate  for a  35mm frame)  while \code{scaling}  maps them  to the
  image size in pixel.  Image dimensions should be odd.

  The        functions       \code{tm.computeDistortionMaps}       and
  \code{tm.distortImage}  are  used   to  simulate  geometrical  image
  distortion due to lenses.  The  meaning of the arguments is the same
  as in the case of  vignetting. Distortion $D = (h'-h)/h$, where $h'$
  is the  radial position  in the distorted  image, is specified  as a
  percentage   (with  a   \code{splinefun}  as   in  the   sample  curve
\code{distagonDistortion}).

  The  function  \code{tm.diffractionKernel}   computes  a  2D  kernel
  representing the Airy function  describing the point spread function
  of an  optical system characterized by the  function arguments.  The
  \code{superSampling} factor  is used to provide  a higher resolution
  kernel: its  elements are no longer  in a one  to one correspondence
  with the original pixels but correspond to a subpixel area.
      
  The  function \code{tm.Lanczos}  provides support  for  kernel based
  image  resampling.   The  sampling  theorem  tells us  that  we  can
  recosntruct  a  properly  sampled  signal  using  the  $\mbox{sinc}$
  function.  This however requires  an infinite series: truncating the
  $\mbox{sinc}$ expansion does not  provide good results, resulting in
  large   Gibbs   oscilaltion.    The   Lanczos  kernel   provides   a
  computationally convenient alternative with good frequency response.
      
  The function \code{tm.photonNoise} provides means to generate photon
  noise  on images.   the quantum  nature of  light, resulting  in its
  description   by   means  of   photons,   the  particles   mediating
  electromagnetic interaction, impacts on the signal to noise ratio of
  the sensors: they see a  discrete flux of particles and the interval
  between  their  arrival follows  Poisson  statistics. The  resulting
  signal variation  is called photon  noise, and for  reasonably large
  flux can be approximated with a Gaussian distribution whose $\sigma$
  is given  by the square root  of the signal.  While  there are other
  sources of noise, some modern  sensors can actually be considered to
  be photon noise limited.

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
  \code{\link{tm.focusBlur}}
}

\references{
Hamilton, Jr., John F., Adams, Jr., James E. (1997).
\emph{Adaptive color plane interpolation in single sensor color electronic camera},
U.S.Patent 5,629,734
}

\examples{
#
# Get sample images
#
tiffdir          <- file.path(system.file(package = "TeMa"),
                              "sampleimages/cameraSimulation")
#
i_ortho          <- ia.scale(ia.readAnimage(file.path(tiffdir,
                                                      "ortho_iso12233_large.tif")),
                             maxValue = 255)
i_orthoRed       <- i_ortho[[1]]
i_orthoRedDetail <- ia.get(i_orthoRed, animask(0,0,1201,256))
#
i_persp          <- ia.scale(ia.readAnimage(file.path(tiffdir,
                                            "persp_iso12233_large.tif")),
                             maxValue = 255)
i_perspRed       <- i_persp[[1]]
i_perspRedDetail <- ia.get(i_perspRed, animask(0,0,1201,256))
#
# Compute the diffraction kernel for a compact camera
#
k_diffraction_f16 <- tm.diffractionKernel(focalLength= 8.9e-3,
                                          fStop=       16,
                                          pixelSize=   (8.8/3264)/1000,
                                          superSampling=3)
#
# and visualize it
#
persp(k_diffraction_f16,
      xlab="x", ylab="y", zlab="Airy function",
      main="Coolpix 8800 diffraction kernel for f16 (3x sampling)")
#
# compute the diffraction kernel for a digital reflex 
#
k_diffraction_f16_D70s <- tm.diffractionKernel(focalLength=50e-3,
                                               fStop=16,
                                               pixelSize=(23.7/3008)/1000,
                                               superSampling = 3)
#
# and note the difference
#
persp(k_diffraction_f16_D70s,
      xlab="x", ylab="y", zlab="Airy function",
      main="D70s diffraction kernel for f16 (3x sampling)")
#
# Now we can see the impact of diffraction on images from the two
# cameras: first the compact one
#
i_perspRedDetail_f16 <-
  ia.convolution(i_perspRedDetail,
                 tm.diffractionKernel(focalLength= 8.9e-3,
                 fStop=       16,
                 pixelSize=   (8.8/3264)/1000))
plot(pixmapGrey(i_perspRedDetail_f16@data),
     main= "Coolpix 8800 blurring effect at f/16")
#
# then the reflex
#
i_perspRedDetail_f16 <-
  ia.convolution(i_perspRedDetail,
                 tm.diffractionKernel(focalLength= 50e-3,
                 fStop=       16,
                 pixelSize=   (23.7/3008)/1000))
plot(pixmapGrey(i_perspRedDetail_f16@data),
     main= "Nikon D70 blurring effect at f/16")
#
# Images are affected by the quantum nature of light: photon noise. Note
# that our synthetic images use the sRGB color space so we first go back
# to a linear space, add the noise, and map again to sRGB. 
#
i_perspRedDetail_f16_iso1600 <-
  ia.apply(i_perspRedDetail_f16,
           function(x, y, z)
              {
               tm.linearToSrgb(
                 tm.photonNoise(tm.srgbToLinear(z/255),
                                gammaCorrection = 1,
                                maxValue        = 1,
                                fullWell        = 48800,
                                fullWellIso     = 100,
                                iso             = 1600)) * 255})
#
plot(pixmapGrey(i_perspRedDetail_f16_iso1600@data),
     main= "Nikon D70 simulation of ISO 1600 at f16")
#
# Now with some examples based on color, illustrating the effect
# of non full RGB imaging
#
i_persp          <- ia.scale(ia.readAnimage(file.path(tiffdir,
                                                      "persp_gretag_large.tif")),
                             maxValue = 255)
i_perspDetail    <- list(ia.get(i_persp[[1]], animask(0,0,1201,256)),
                         ia.get(i_persp[[2]], animask(0,0,1201,256)),
                         ia.get(i_persp[[3]], animask(0,0,1201,256)))
#
i_perspDetail_bayerized <- tm.bayerizeImage(i_perspDetail)
#
plot(pixmapGrey(i_perspDetail_bayerized@data),
     main="Bayerized version of a full RGB image")
#
i_perspDetail_debayerized <- tm.debayerizeImage(i_perspDetail_bayerized)
#
plot(pixmapGrey(i_perspDetail_debayerized[[1]]@data - i_perspDetail[[1]]@data),
     main="De-bayerized image error")
#
}

\keyword{array}
