\name{tm.histogramEqualize}

\alias{tm.histogram}
\alias{tm.cumulateHistogram}
\alias{tm.ehUniform}
\alias{tm.ehExponential}
\alias{tm.ehGaussian}
\alias{tm.ehRayleigh}
\alias{tm.ehCubeRoot}
\alias{tm.ehLogarithmic}

\title{Histogram equalization}

\description{

  Histogram equalization transforms image intensity values in order
  to improve image readability or to normalize the distributions arising
  from different acquisition chains to a common standard setting.

}

\usage{
tm.histogram(img, mask = FALSE, srcRange = c(0,255), dstRange =
c(0,255))
tm.cumulateHistogram(histo)

tm.ehUniform(f, eqSpecs = list(gMin=0, gMax=255))
tm.ehExponential(f, eqSpecs = list(gMin=0, alpha=0.01))
tm.ehRayleigh(f, eqSpecs = list(gMin=0, alpha=50))
tm.ehCubeRoot(f, eqSpecs = list(gMin=0, gMax=256))
tm.ehLogarithmic(f, eqSpecs = list(gMin=1, gMax=256))
tm.ehGaussian(f, eqSpecs = list(gMin=0, gMax=256, mu=128, sd=30))

tm.histogramEqualizeMap(histo,
                        eqSpecs =  list(type="Gaussian",
                                         gMin=0,gMax=256, mu=128, sd=30))

tm.histogramEqualize(img, eqSpecs = list(type="Gaussian",
                                         gMin=0,gMax=256, mu=128, sd=30))
}

\arguments{
  \item{img}{a \code{animage}}
  \item{srcRange}{a two elements vector \code{c(m, M)} used to specify
	intensity value stretching:
	image values less (greater) than \code{m} (\code{M}) are set to
	\code{m} (\code{M})}
  \item{f}{a real number in [0,1] specifying a normalized cumulative
	histogram value}
  \item{dstRange}{a two elements vector \code{c(m, M)} used to specify
	intensity value stretching: values from \code{srcRange} are
	translated and scaled to fit \code{dstRange}}
  \item{eqSpecs}{a \code{list} of named elements providing a full
	specfication of the required equalization function}
  \item{histo}{a \code{list} composed of two named elements:
	\code{hdata}, a 1D array providing the counts for each image value,
	\code{hoffset}, identifying the intensity value corresponding to the
	first element of \code{hdata}}

}

\value{

  The function \code{tm.histogram} and \code{tm.cumulativeHistogram} 
  return respectively an histogram and its cumulative version. The
  functions whose name starts with \code{tm.eh} return a real value
  representing the target image intensity value corresponding to the
  original  value  whose  corresponding  cumulative histogram  value  is
  \code{f}. The function \code{tm.histogramEqualizeMap} returns a look
  up table whose format is a list with two named fields, \code{lutOffset} and
  \code{lutData}.   The function  \code{tm.histogramEqualize}  returns a
  \code{animage}.

  
}

\details{

  Histogram equalization functions are meant to operate on integer
  \code{animage}s: real \code{animage}s are converted to integer ones
  based on the specified \code{srcRange} and \code{dstRange}
  provided to \code{tm.histogram}. The implementation is based on the
  computation of a lookup table mapping the cumulative histogram
  (computed by \code{tm.cumulateHistogram}) to the required target values.
  Several maps are available (functions \code{tm.eh....}) including the
  possibility of specifying a target histogram.

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\references{
  Pratt, W.K. (2007)
  \emph{Digital Image Processing},
  Wiley-Interscience
}

\examples{
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages/")
imgA         <- ia.integerAnimage(
                 as.animage(
                  getChannels(
                   read.pnm(file.path(sampleimages,"sampleFace_07.pgm")))*255))
imgB         <- ia.integerAnimage(
                 as.animage(
                  getChannels(
                   read.pnm(file.path(sampleimages,"sampleFace_08.pgm")))*255))
#
hA                <- tm.histogram(imgA)
hmGaussian        <- tm.histogramEqualizeMap(hA,
                                             eqSpecs = list(type="Gaussian",
                                                         Gmin=0, gMax=256, mu=128, sd=30))
imgAeqGaussian    <- ia.table(imgA, hmGaussian$lutData, tableOffset=hmGaussian$lutOffset)
#
imgAeqGaussian    <- tm.histogramEqualize(imgA,
                                               eqSpecs = list(type="Gaussian",
                                                           gMin=0, gMax=256, mu=128, sd=30))
imgAeqRayleigh    <- tm.histogramEqualize(imgA, list(type="Rayleigh", gMin=0, alpha=50))
imgAeqUniform     <- tm.histogramEqualize(imgA, list(type="uniform", gMin=32, gMax=224))
imgAeqExponential <- tm.histogramEqualize(imgA, list(type="exponential", gMin=0, alpha=0.01))
imgAeqCubeRoot    <- tm.histogramEqualize(imgA, list(type="cubeRoot", gMin=0, gMax=256))
imgAeqLogarithmic <- tm.histogramEqualize(imgA, list(type="logarithmic", gMin=1, gMax=256))
#
oldpar        <- par(no.readonly = TRUE)
#
par(mfrow   = c(2,3),
    mar     = c(2,1,0,1),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1.5)
#
plot(pixmapGrey(imgAeqUniform@data),     xlab="uniform")
plot(pixmapGrey(imgAeqGaussian@data),    xlab="Gaussian")
plot(pixmapGrey(imgAeqRayleigh@data),    xlab="Rayleigh")
#
par(mar     = c(4,4,0,1),
    mgp     = c(2,0.5,0.0),
    lwd     = 0.5)
#
hist(imgAeqUniform@data,  xlab = "", main="", probability = TRUE)
hist(imgAeqGaussian@data, xlab = "", ylab = "",  main="", probability = TRUE)
hist(imgAeqRayleigh@data, xlab = "", ylab = "",  main="", probability = TRUE)
#
par(mar     = c(2,1,0,1),
    mgp     = c(1,1,1))
#
plot(pixmapGrey(imgAeqCubeRoot@data),    xlab="cubeRoot")
plot(pixmapGrey(imgAeqLogarithmic@data), xlab="logarithmic")
plot(pixmapGrey(imgAeqExponential@data), xlab="exponential")
#
par(mar     = c(4,4,0,1),
    mgp     = c(2,0.5,0.0),
    lwd     = 0.5)
#
hist(imgAeqCubeRoot@data,    xlab = "", main="", probability = TRUE)
hist(imgAeqLogarithmic@data, xlab = "", ylab = "",  main="", probability = TRUE)
hist(imgAeqExponential@data, xlab = "", ylab = "",  main="", probability = TRUE)
#
par(oldpar)
#
imgBeqTarget      <- tm.histogramEqualize(imgB,
                                          list(type="target", histo=tm.histogram(imgA)))
#
par(mfrow   = c(2,3),
    mar     = c(2,1,0,1),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1.5)
#
plot(pixmapGrey(imgA@data),            xlab="A")
plot(pixmapGrey(imgB@data),            xlab="B")
plot(pixmapGrey(imgBeqTarget@data),    xlab="B with A target")
#
par(mar     = c(4,4,0,1),
    mgp     = c(2,0.5,0.0),
    lwd     = 0.5)
#
hist(imgA@data,         xlab = "", main="", probability = TRUE)
hist(imgB@data,         xlab = "", ylab = "",  main="", probability = TRUE)
hist(imgBeqTarget@data, xlab = "", ylab = "",  main="", probability = TRUE)
#
par(mfrow   = c(1,2),
    mar     = c(4,4,0,1),
    mgp     = c(2,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1.5)
#
plot(imgA@data/255,imgB@data/255,
     xlab="A",ylab="B",xlim=c(0,1),
     ylim=c(0,1), lwd=0.1)
#
plot(imgAeqTarget@data/255,imgB@data/255,
     xlab="A",ylab="B",xlim=c(0,1),
     ylim=c(0,1), lwd=0.1)
#
}

\keyword{array}



