#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arrayTools/aniCommon.h>


#ifndef MAX
#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#endif

#ifndef MIN
#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#endif

#ifndef ABS
#define ABS(x)    (((x) > 0)   ? (x) : (-(x)))
#endif

float circleOfConfusion(float focalLength,    // in meters
                        float focalDistance,  // in meters
                        float fStop,          // 2.8 ...
                        float depth)          // in meters
{
  float coc;

  coc = ((focalLength/fStop) * (focalDistance * focalLength) / 
         (focalDistance - focalLength) *
         fabs(1/depth - 1/focalDistance));

  return(coc);
}

extern "C"  {

  // We must work on real arrays to ensure enough precision (right?)
  // and we expect the depthMap to be metric in millimeters (i.e. 
  // if it scaling was relative, it must have been properly
  // normalized to the actual metric distances)
  // 
  SEXP tmFocusBlur1D(SEXP channels,
					 SEXP depthMap,
					 SEXP focalLength,
					 SEXP fStop,
					 SEXP focalDistance,
					 SEXP scale,     // the scale factor to map onto the sensor
					 SEXP maxRadius) // above which nop
  {
    SEXP    tm_r; // will contain the list of defocused channels
    int     x0, y0, w, h;
    double  o;

    // 1. determine the number of channels
    //
    int    numberOfChannels  = LENGTH(channels);

    // 2. create as many new (protected) images as provided channels
    // 
    //    we need to get some information from the first available
    //    channel
    //
    x0 = GET_ANIMAGE_X(VECTOR_ELT(channels,0));
    y0 = GET_ANIMAGE_Y(VECTOR_ELT(channels,0));
    w  = GET_ANIMAGE_W(VECTOR_ELT(channels,0));
    h  = GET_ANIMAGE_H(VECTOR_ELT(channels,0));
    o  = GET_ANIMAGE_OR(VECTOR_ELT(channels,0));
    
    //   protect the return list result
    //
    PROTECT(tm_r = NEW_LIST(numberOfChannels));

    //   and add the required number of channels
    //
    double ***_dch            = new double ** [numberOfChannels];
    double ***_fch            = new double ** [numberOfChannels];
    double  **_dp             = new double  * [w];
    SEXP   *defocusedChannels = new SEXP      [numberOfChannels];

    //    precompute the depth pointers for more flexible read out
    //
    _dp[0] = GET_ANIMAGE_DR(depthMap);

    for(int x = 1; x < w; x++)
      _dp[x] = _dp[x-1] + h;

    for(int ch = 0; ch < numberOfChannels; ch++)
      {
        // create the new channels
        //
        SET_ELEMENT(tm_r, ch, (defocusedChannels[ch]=newAnimage(x0,y0,w,h,o)));

        // fill in pointer information
        //
        {
          _dch[ch]    = new double * [w];
          _dch[ch][0] = GET_ANIMAGE_DR(defocusedChannels[ch]);

          for(int x = 1; x < w; x++)
            _dch[ch][x] = _dch[ch][x-1] + h;
        }
        // and for the original pin hole channels
        //
        {
          _fch[ch]    = new double * [w];
          _fch[ch][0] = GET_ANIMAGE_DR(VECTOR_ELT(channels, ch));

          for(int x = 1; x < w; x++)
            _fch[ch][x] = _fch[ch][x-1] + h;
        }
      }
    
    // 3. compute a map of the circle of confusion (and of its 1/area)
    //    for each pixel position
    //
    double fL = REAL(focalLength)[0];
    double fD = REAL(focalDistance)[0];
    double fS = REAL(fStop)[0];
    double s  = REAL(scale)[0];
    int    mR = int(REAL(maxRadius)[0]+0.5); // the width of the filter 


    ia::array<float> coc(w, h);             // the radius of the coc
    ia::array<float> nct(w, h);             // and its area^-1
    ia::array<float> tch[numberOfChannels]; // buffers for the double pass 

    for(int ch = 0; ch < numberOfChannels; ch++)
      tch[ch] = (* new ia::array<float>(w,h)).temporary();

    for(int x = 0; x < w; x++) 
      {
        for(int y = 0; y < h; y++)
          {
            coc[x][y] = s * circleOfConfusion(fL, fD, fS, _dp[x][y]);
            nct[x][y] = 1.0 / (M_PI * coc[x][y] * coc[x][y]);
          }
      }

    // 4. We apply the approach presented in 
    //    'Accurate Depth of Field Simulation in Real Time', 
	//    Zhou T, Chen JX and Pullen M
    //    COMPUTER GRAPHICS forum 26:15-23 2007
    //
    //    two pass filtering:
    //    
    //      the vertical pass
    //
    float             Op, Ip, Lp, Wn;
    ia::vector<float> W(-mR, mR);

    for(int x = 0; x < w; x++) 
      {
        for(int y = 0; y < h; y++)
          {
            int r0   = - MIN(y,     mR);
            int r1   = MIN(h-1-y, mR);

            W  = 0.0;   // reset weights
            Wn = 0.0;   // reset normalization

            for(int r = r0, yy = y+r0; r <= r1; r++, yy++)
              {
                float cr = coc[x][yy];

                // compute the overlap factor
                //
                if(cr <= ABS(r))
                  Op = 0.0;
                else if (cr >= ABS(r) + 1.0)
                  Op = 1.0;
                else 
                  Op = cr - ABS(r);

                Ip    = nct[x][yy];                            // intensity
                Lp    = (_dp[x][yy] > fD) ? MIN(cr,1.0) : 1.0; // leakage
                W[r]  = (r == 0) ? Ip : Op * Ip * Lp;          // weight
                Wn   += W[r];                                  // normalization
              }

            // now proceed to compute the defocused channels
            //
            for(int ch = 0; ch < numberOfChannels; ch++)
              {
                float value = 0;

                for(int r = r0, yy = y+r0; r <= r1; r++, yy++)
                    value += W[r] * _fch[ch][x][yy];

                // fprintf(stderr, "%f %f\n", Wn, coc[x][y]);

                if(Wn > 0.0)
                  tch[ch][x][y] = value / Wn;
                else
                  tch[ch][x][y] = 0.0;
              }
          }
      }

    //    
    //    and the horizontal one
    //
    for(int y = 0; y < h; y++)
      {
        for(int x = 0; x < w; x++) 
          {
            int r0   = - MIN(x,  mR);
            int r1   = MIN(w-1-x, mR);

            W  = 0.0; // reset weights
            Wn = 0;   // reset normalization

            for(int r = r0, xx = x+r0; r <= r1; r++, xx++)
              {
                float cr = coc[xx][y];

                // compute the overlap factor
                //
                if(cr <= ABS(r))
                  Op = 0.0;
                else if (cr >= ABS(r) + 1.0)
                  Op = 1.0;
                else 
                  Op = cr - ABS(r);

                Ip    = nct[xx][y];                            // intensity
                Lp    = (_dp[xx][y] > fD) ? MIN(cr,1.0) : 1.0; // leakage
                W[r]  = (r == 0) ? Ip : Op * Ip * Lp;          // weight
                Wn   += W[r];                                  // normalization
              }

            // now proceed to compute the defocused channels
            //
            for(int ch = 0; ch < numberOfChannels; ch++)
              {
                float value = 0;

                for(int r = r0, xx = x+r0; r <= r1; r++, xx++)
                    value += W[r] * tch[ch][xx][y];

                _dch[ch][x][y] = value / Wn;
              }
          }
      }

    // 5. clean up and return a list with the new channels
    //
        for(int ch = 0; ch < numberOfChannels; ch++)
          {
            delete _dch[ch];
            delete _fch[ch];
          }

        delete _dch;
        delete _fch;
        delete _dp;


        delete defocusedChannels;

    UNPROTECT(1);
    
    return(tm_r);
  }

  SEXP tmFocusBlur2D(SEXP channels,
					 SEXP depthMap,
					 SEXP focalLength,
					 SEXP fStop,
					 SEXP focalDistance,
					 SEXP scale,     // the scale factor to map onto the sensor
					 SEXP maxRadius) // above which nop
  {
    SEXP    tm_r; // will contain the list of defocused channels
    int     x0, y0, w, h;
    double  o;

    // 1. determine the number of channels
    //
    int    numberOfChannels  = LENGTH(channels);

    // 2. create as many new (protected) images as provided channels
    // 
    //    we need to get some information from the first available
    //    channel
    //
    x0 = GET_ANIMAGE_X(VECTOR_ELT(channels,0));
    y0 = GET_ANIMAGE_Y(VECTOR_ELT(channels,0));
    w  = GET_ANIMAGE_W(VECTOR_ELT(channels,0));
    h  = GET_ANIMAGE_H(VECTOR_ELT(channels,0));
    o  = GET_ANIMAGE_OR(VECTOR_ELT(channels,0));
    
    //   protect the return list result
    //
    PROTECT(tm_r = NEW_LIST(numberOfChannels));

    //   and add the required number of channels
    //
    double ***_dch            = new double ** [numberOfChannels];
    double ***_fch            = new double ** [numberOfChannels];
    double  **_dp             = new double  * [w];
    SEXP   *defocusedChannels = new SEXP      [numberOfChannels];

    //    precompute the depth pointers for more flexible read out
    //
    _dp[0] = GET_ANIMAGE_DR(depthMap);

    for(int x = 1; x < w; x++)
      _dp[x] = _dp[x-1] + h;

    for(int ch = 0; ch < numberOfChannels; ch++)
      {
        // create the new channels
        //
        SET_ELEMENT(tm_r, ch, (defocusedChannels[ch]=newAnimage(x0,y0,w,h,o)));

        // fill in pointer information
        //
        {
          _dch[ch]    = new double * [w];
          _dch[ch][0] = GET_ANIMAGE_DR(defocusedChannels[ch]);

          for(int x = 1; x < w; x++)
            _dch[ch][x] = _dch[ch][x-1] + h;
        }
        // and for the original pin hole channels
        //
        {
          _fch[ch]    = new double * [w];
          _fch[ch][0] = GET_ANIMAGE_DR(VECTOR_ELT(channels, ch));

          for(int x = 1; x < w; x++)
            _fch[ch][x] = _fch[ch][x-1] + h;
        }
      }
    
    // 3. compute a map of the circle of confusion (and of its 1/area)
    //    for each pixel position
    //
    double fL = REAL(focalLength)[0];
    double fD = REAL(focalDistance)[0];
    double fS = REAL(fStop)[0];
    double s  = REAL(scale)[0];
    int    mR = int(REAL(maxRadius)[0]+0.5); // the width of the filter 


    ia::array<float> coc(w, h);             // the radius of the coc
    ia::array<float> nct(w, h);             // and its area^-1
    ia::array<float> tch[numberOfChannels]; // buffers for the double pass 

    for(int ch = 0; ch < numberOfChannels; ch++)
      tch[ch] = (* new ia::array<float>(w,h)).temporary();

    for(int x = 0; x < w; x++) 
      {
        for(int y = 0; y < h; y++)
          {
            coc[x][y] = s * circleOfConfusion(fL, fD, fS, _dp[x][y]);
            nct[x][y] = 1.0 / (M_PI * coc[x][y] * coc[x][y]);
          }
      }

    // 4. We apply the approach presented in 
    //    'Accurate Depth of Field Simulation in Real Time', 
	//    Zhou T, Chen JX and Pullen M
    //    COMPUTER GRAPHICS forum 26:15-23 2007
    //
    //    but modified to a single 2D pass that is 
	//    slower but more accurate
    //
    double            Op, Ip, Lp, Wn;
    ia::array<double> W(-mR, mR,-mR,mR);

    for(int x = 0; x < w; x++) 
      {
        for(int y = 0; y < h; y++)
          {
            int ry0   = - MIN(y,     mR);
            int ry1   =   MIN(h-1-y, mR);
            int rx0   = - MIN(x,     mR);
            int rx1   =   MIN(w-1-x, mR);

            Wn = 0.0;   // reset normalization

            for(int ry = ry0, yy = y+ry0; ry <= ry1; ry++, yy++)
			  for(int rx = rx0, xx = x+rx0; rx <= rx1; rx++, xx++)
				{
				  double  R = sqrt(ry*ry+rx*rx);
				  double cr = coc[xx][yy];

				  // compute the overlap factor
				  //
				  if(cr <= R)
					Op = 0.0;
				  else if (cr >= R + 1.0)
					Op = 1.0;
				  else 
					Op = cr - R;

				  Ip         = nct[xx][yy];                            // intensity
				  Lp         = (_dp[xx][yy] > fD) ? MIN(cr,1.0) : 1.0; // leakage
				  W[ry][rx]  = ((ry == 0) && (rx == 0)) ? Ip : Op * Ip * Lp;           // weight
				  Wn        += W[ry][rx];                              // normalization
				}

            // now proceed to compute the defocused channels
            //
            for(int ch = 0; ch < numberOfChannels; ch++)
              {
                float value = 0;

                for(int ry = ry0, yy = y+ry0; ry <= ry1; ry++, yy++)
				  for(int rx = rx0, xx = x+rx0; rx <= rx1; rx++, xx++)
                    value += W[ry][rx] * _fch[ch][xx][yy];

                // fprintf(stderr, "%f %f\n", Wn, coc[x][y]);

                if(Wn > 0.0)
                  _dch[ch][x][y] = value / Wn;
                else
                  _dch[ch][x][y] = 0.0;
              }
          }
      }

    // 5. clean up and return a list with the new channels
    //
        for(int ch = 0; ch < numberOfChannels; ch++)
          {
            delete _dch[ch];
            delete _fch[ch];
          }

        delete _dch;
        delete _fch;
        delete _dp;


        delete defocusedChannels;

    UNPROTECT(1);
    
    return(tm_r);
  }

}
