#
# \litcode{A simple approach to robustified correlation}
#
# Pearson correlation coefficient is closely related to the computation
# of the variance of a set of numbers: it is the dot product of two vectors
# previously normalized to zero average and unitary variance. The sensitivity
# of the standard sample variance estimator to the presences of outlying values,
# such as those due to salt and pepper noise, affects the robustness of the
# Pearson correlation coefficient. Equation~\TM:4.17 shows how we can
# define a robust correlation coefficient using a robust scale estimator.
# The function \code{tm.robustifiedCorrelation} implements Equation~\TM:4.17 using
# the scale estimates provided by \code{tm.robustEstimates}.
#
tm.robustifiedCorrelation <- function(X, Y, mode="O") {
  #
  # \code{X} and \code{Y} represent two images, usually normalized
  # to zero location and unitary scale using \code{tm.normalizeImage}.
  # The third parameter, \code{mode}, can assume three different values:
  # \code{"p"}, corresponding to the usual (non robust) sample estimator,
  # \code{"M"}, corresponding to a (non-refined) \code{tanh} estimator, and
  # \code{"O"}, corresponding to a one-step version of the \code{tanh} estimator.
  #
  # In order to exploit Equation~\TM:4.17, we must generate two new random
  # variables that correspond respectively to the sum and to the difference
  # of the two images:
  #
  XpY <- ia.add(X, Y)
  XmY <- ia.sub(X, Y)
  #
  # proceeding to the computations required by  Equation~\TM:4.17 
  #
  sp  <- tm.robustEstimates(array(XpY@data, dim=c(ia.size(XpY),1)), mode)
  sm  <- tm.robustEstimates(array(XmY@data, dim=c(ia.size(XmY),1)), mode)
  #
  (sp[2]*sp[2] - sm[2]*sm[2])/ (sp[2]*sp[2] + sm[2]*sm[2])
}
