#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arrayTools/Array.h>
#include <arrayTools/PpmIO.h>
#include <arrayTools/JpgIO.h>
#include <arrayTools/TiffIO.h>
#include <vector>
#include <unistd.h>
#include <tiffio.h>


// To each non empty histogram bin we associate the corresponding value
// of the cumulative distribution (we need this to break ties via 
// scan order position)
//
void rankHistogram(ia::vector<int> &H,
                   ia::vector<int> &R)
{
  int r = 0;

  for(int i = 0; i < 256; i++)
    if(H[i] > 0) 
      {
        r    += H[i];
        R[i]  = r;
      }
}
//
// Compute the Bhat-Nayar distance of two ranked templates.
// The first template (reference) is represented with the positions
// of its sorted pixels, the second is a standard image (2D)
// where each pixel is assigned its rank.
//
// Alpha represents the relative weight of the average distance with
// respect to the standard (max) one: 
//
// alpha = 0 -> max
// alpha = 1 -> average
//
float compareRankedTemplates(ia::array<int> &iA,
                             ia::array<int> &B,
                             float           alpha)
{
  int              n = B.rows() * B.cols();
  int              d = 0;
  float            D = 0.0, S = 0.0, aD = 0.0;
  ia::vector<int> H(0,n);

  for(int i = 0; i < n; i++)
    {
      int r = B[iA[i][1]][iA[i][0]];

      H[r] += 1;
      d    += (r > i) ? 1 : 0;
      d    -= H[i];
      
      if(d > D)
        D = d;

      aD += d;
    }

  S = 1 - ((1-alpha)*D/(n/2) + alpha*aD/((n*n)/4));

  return(S);
}
//
// Efficient ordinal matching exploiting overlapping regions
// histogram update
//
void scanningHistogramMatch(ia::array<u_char> &I,
                            ia::array<u_char> &A,
                            float              alpha)
{
  int               tw = A.cols();
  int               th = A.rows();
  int               n  = tw * th;
  int               x0, x1, y0, y1;
  ia::vector<int>   H0(0,255);  // histogram from beginning of previous line
  ia::vector<int>   Hc(0,255);  // current histogram
  ia::vector<int>   R(0,255);   // ranks for each histogram value
  ia::vector<int>   Rc(0,255);  // ranks for each histogram value
  ia::array<int>    B(0,th-1,0,tw-1);
  ia::array<int>    T(0,th-1,0,tw-1);

  //
  // we must first prepare the template for the comparison 
  // (A -> B)
  //
  // get image dimensions
  //
  A.stat(x0, x1, y0, y1);
  //
  // compute histogram 
  //
  for(int y = y0; y <= y1; y++)
    for(int x = x0; x <= x1; x++)
      Hc[A[y][x]] += 1;
  //
  // rank it
  //
  rankHistogram(Hc, Rc);
  //
  // and create the ordinal template, breaking ties
  // using scan order
  //
  for(int y = y0, j=0; y <= y1; y++, j++)
    for(int x = x0, i=0; x <= x1; x++, i++)
      {
        B[j][i]      = Rc[A[y][x]] - 1;
        Rc[A[y][x]] -= 1;
      }
  //
  // We need to map progressive ranks into
  // the pixels of the reference for efficient
  // comparison
  //
  ia::array<int> iB(n,2);
  //
  for(int j=0; j < th; j++)
    for(int i=0; i < tw; i++)
      {
        iB[B[j][i]][0] = i;
        iB[B[j][i]][1] = j;
      }
  //
  // and switch to the complete image
  //
  // get image dimensions
  //
  I.stat(x0, x1, y0, y1);
  //
  // compute limits for actual computation
  //
  x1 -= tw;
  y1 -= th;
  //
  // allocate result matrix
  //
  ia::array<float> BN(y0,y1,x0,x1);
  //
  // compute first histogram 
  //
  for(int y = y0; y < y0+th; y++)
    for(int x = x0; x < x0+tw; x++)
      H0[I[y][x]] += 1;
  //
  // and scan the whole image
  //
  float minD = n;
  float maxD = 0;
  //
  for(int y = y0; y <= y1; y++)
    {
      if(y > y0) 
        {
          // we must update H0:
          //
          // we remove previous upper row
          // and add the new entry (lowest)
          //
          for(int i = 0; i < tw; i++)
            {
              H0[I[y-1][x0+i]]--;
              H0[I[y+th-1][x0+i]]++;
            }
        }
      // 
      // and we make it current
      //
      Hc = H0;
      //
      for(int x = x0; x <= x1; x++)
        {
          if(x > x0) 
            {
              // we must update Hc
              //
              for(int j = 0; j < th; j++)
                {
                  Hc[I[y+j][x-1]]--;
                  Hc[I[y+j][x+tw-1]]++;
                }
            }
          //
          // and rank its entries
          //
          rankHistogram(Hc, Rc);
          //
          // compute the sorted window
          //
          for(int yy = y, j=0; yy < y+th; yy++, j++)
            for(int xx = x, i=0; xx < x+tw; xx++, i++)
              {
                T[j][i]        = Rc[I[yy][xx]] - 1;
                Rc[I[yy][xx]] -= 1;
              }
          //
          // and proceed to match it to reference
          //
          BN[y][x] = compareRankedTemplates(iB, T, alpha);

          if(BN[y][x] < minD)
			{
			  minD = BN[y][x];
			}
        }
    }
}
//
// Sample usage:
//
// ./bnCorrelation RobertoNoisy.pgm RobertoEye.pgm 0.5
//
//
int main(int argc,
         char **argv)
{
  ia::array<u_char> I;
  ia::array<u_char> T;

  // read image
  //
  ReadPgm(I, argv[1]);
  //
  // and template
  //
  ReadPgm(T, argv[2]);
  //
  //
  float alpha = atof(argv[3]);
  //
  if(alpha > 1)
    alpha = 1;
  else if (alpha < 0)
    alpha = 0;
  //
  scanningHistogramMatch(I, T, alpha);
  //
}
  
